import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;

public class AddressBook implements Serializable{
	//initializing the address book
	private SortedMap<String, Contact> contacts = new TreeMap<String, Contact>();
	
	/**
	 * adds the contact c to the address book
	 * @param c - a new contact to add to the address book
	 * @throws ContactAlreadyExistsException - a contact with the same name
	 * (case insensitive) already exists
	 * @post this.containsValue(c)
	 */
	public void add(Contact c) throws ContactAlreadyExistsException {
		String key = c.getName().toLowerCase();
		if (contacts.containsKey(key))
			throw new ContactAlreadyExistsException();
		contacts.put(key, c);
	}
	
	/**
	 * returns the contact with the given name from the address book
	 * @param name - the name of the contact we wish to get
	 * @return the contact with the name given
	 * @throws EmptyNameException - name given is null/empty
	 * @throws ContactDoesNotExistsException - the address book does not contains
	 * a contact with the given name
	 */
	public Contact get(String name) throws EmptyNameException,
											ContactDoesNotExistsException {
		if ((name == null) || (name == ""))
			throw new EmptyNameException();
		if (!contacts.containsKey(name.toLowerCase()))
			throw new ContactDoesNotExistsException();
		return contacts.get(name.toLowerCase());
	}
	
	/**
	 * deletes the contact with the given name from the address book
	 * @param name - the name of the contact we wish to delete
	 * @throws EmptyNameException - name given is null/empty
	 * @throws ContactDoesNotExistsException - the address book does not contains
	 * a contact with the given name
	 * @post !this.myContacts.containsKey(name.toLowerCase())
	 */
	public void delete(String name) throws EmptyNameException,
											ContactDoesNotExistsException {
		if ((name == null) || (name == ""))
			throw new EmptyNameException();
		if (!contacts.containsKey(name.toLowerCase()))
			throw new ContactDoesNotExistsException();
		contacts.remove(name.toLowerCase());
	}
	
	/**
	 * modifies the contact in the address book with the same name as c to c
	 * @param c - the contact to modify
	 * @throws ContactDoesNotExistsException - the given contacts name doesn't
	 * exists in any contact in the address book  
	 */
	public void modify(Contact c) throws EmptyNameException,
											ContactDoesNotExistsException,
											ContactAlreadyExistsException {
		//EmptyNameException - will never be thrown since it sends an initialized contact
		//ContactAlreadyExistsException - will never be thrown since it come right after
		//the old contact with the same name has been deleted from the address book
		this.delete(c.getName());
		this.add(c);
	}
	
	/**
	 * returns an iterator for the address book
	 * @return contacts.values().iterator()
	 */
	public Iterator<Contact> getContacts(){
		return contacts.values().iterator();
	}
	
	/**
	 * returns the size of the address book
	 * @return contacts.size()
	 */
	public int getCount(){
		return contacts.size();
	}
	
	/**
	 * returns an iterator for the address book, only for the contacts with names
	 * starting with the given prefix (case insensitive)
	 * @param prefix - the prefix to check the contacts by
	 * @return an iterator by the rules above
	 */
	public Iterator<Contact> search(String prefix){
		if ((prefix.equals(new String(""))) || (prefix == null)) return getContacts();
		else return new prefixIterator(prefix, getContacts());
	}
	
	/**
	 * saves the contacts to a file named filename
	 * @param filename - the name of the file to save
	 * @throws IOException - in case of any input/output error
	 */
	 public void save(String filename) throws IOException {
		try{
			FileOutputStream output = new FileOutputStream(filename);
			ObjectOutputStream out = new ObjectOutputStream(output);
			out.writeObject(this.contacts);
			out.close();
		} catch (IOException e){
			throw new IOException();
		}
	 }
	 
	 /**
	  * loads a contacts file to replace this.contacts
	  * @param filename - the filename to read from
	  * @throws Exception - in case of any input/output error
	  */
	 public void load(String filename) throws Exception {
		 try{
			 FileInputStream input = new FileInputStream(filename);
			 ObjectInputStream in = new ObjectInputStream(input);
			 this.contacts = (TreeMap<String, Contact>) in.readObject();
			 in.close();
		 } catch (Exception e){
			 throw new IOException();
		 }
	 }
}
